'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
/**
 * datastructures-js/avl-tree
 * @copyright 2018 Eyas Ranjous <eyas.ranjous@gmail.com>
 * @license MIT
 */

/**
 * avl tree node
 * @function
 */
var node = function node(v, p, l, r) {
  var value = v;
  var parent = p || null;
  var left = l || null;
  var right = r || null;
  var height = 1;

  /**
   * @param {object} value
   */
  var setValue = function setValue(val) {
    value = val;
  };

  /**
   * @returns {string|number}
   */
  var getValue = function getValue() {
    return value;
  };

  /**
   * @param {object} pr
   */
  var setParent = function setParent(pr) {
    parent = pr;
  };

  /**
   * @returns {object} node
   */
  var getParent = function getParent() {
    return parent;
  };

  /**
   * @param {BinaryNode} node
   */
  var setLeft = function setLeft(lf) {
    left = lf;
  };

  /**
   * @returns {object} node
   */
  var getLeft = function getLeft() {
    return left;
  };

  /**
   * @param {object} rg
   */
  var setRight = function setRight(rg) {
    right = rg;
  };

  /**
   * @returns {BinaryNode}
   */
  var getRight = function getRight() {
    return right;
  };

  /**
   * @param {number} height
   */
  var setHeight = function setHeight(h) {
    height = h;
  };

  /**
   * @returns {number}
   */
  var getHeight = function getHeight() {
    return height;
  };

  // avl tree node api
  return {
    setValue: setValue,
    getValue: getValue,
    setParent: setParent,
    getParent: getParent,
    setRight: setRight,
    getRight: getRight,
    setLeft: setLeft,
    getLeft: getLeft,
    setHeight: setHeight,
    getHeight: getHeight
  };
};

/**
 * binary search tree
 * @function
 */
var avlTree = function avlTree() {
  var rootNode = null;
  var nodesCount = 0;

  /**
   * @returns {object} node
   */
  var root = function root() {
    return rootNode;
  };

  /**
   * @returns {number}
   */
  var count = function count() {
    return nodesCount;
  };

  /**
   * gets max value node in the tree
   * @returns {object} node
   */
  var max = function max(startingNode) {
    var currentNode = startingNode || rootNode;
    while (currentNode !== null && currentNode.getRight() !== null) {
      currentNode = currentNode.getRight();
    }
    return currentNode;
  };

  /**
   * gets min value node in the tree
   * @returns {object} node
   */
  var min = function min(startingNode) {
    var currentNode = startingNode || rootNode;
    while (currentNode !== null && currentNode.getLeft() !== null) {
      currentNode = currentNode.getLeft();
    }
    return currentNode;
  };

  /**
   * gets the height of a node
   * @param {object} n - avl node
   * @returns {number}
   */
  var height = function height(n) {
    return n && n.getHeight() || 0;
  };

  /**
   * updates the height of a node
   * @param {object} n - avl node
   */
  var updateHeight = function updateHeight(n) {
    if (n !== null) {
      var rightHeight = height(n.getRight());
      var leftHeight = height(n.getLeft());
      n.setHeight(Math.max(rightHeight, leftHeight) + 1);
    }
  };

  /**
   * calculates the balance of a node based on its childrens heights
   * @returns {number}
   */
  var calculateBalance = function calculateBalance(n) {
    if (n !== null) {
      return height(n.getLeft()) - height(n.getRight());
    }
    return 0;
  };

  /**
   * performs a left rotation of a node (counter-clockwise)
   * @param {object} n - the node to be rotated
   */
  var rotateLeft = function rotateLeft(n) {
    var parent = n.getParent();
    var right = n.getRight();
    var rightLeft = right.getLeft();
    n.setRight(rightLeft);
    right.setLeft(n);
    if (n === rootNode) {
      rootNode = right;
    } else if (parent.getValue() < right.getValue()) {
      parent.setRight(right);
    } else if (parent.getValue() > right.getValue()) {
      parent.setLeft(right);
    }
    if (rightLeft !== null) {
      rightLeft.setParent(n);
    }
    n.setParent(right);
    right.setParent(parent);
    updateHeight(n);
    updateHeight(n.getParent());
  };

  /**
   * performs a right rotation of a node (clockwise)
   * @param {object} n - avl node
   */
  var rotateRight = function rotateRight(n) {
    var parent = n.getParent();
    var left = n.getLeft();
    var leftRight = left.getRight();
    n.setLeft(leftRight);
    left.setRight(n);
    if (n === rootNode) {
      rootNode = left;
    } else if (parent.getValue() > left.getValue()) {
      parent.setLeft(left);
    } else if (parent.getValue() < left.getValue()) {
      parent.setRight(left);
    }
    if (leftRight !== null) {
      leftRight.setParent(n);
    }
    n.setParent(left);
    left.setParent(parent);
    updateHeight(n);
    updateHeight(n.getParent());
  };

  /**
   * performs a left rotation of left child then a right rotation of node
   * @param {object} n - avl node
   */
  var rotateLeftRight = function rotateLeftRight(n) {
    rotateLeft(n.getLeft());
    rotateRight(n);
  };

  /**
   * performs a right rotation of right child then a left rotation of node
   * @param {object} n - avl node
   */
  var rotateRightLeft = function rotateRightLeft(n) {
    rotateRight(n.getRight());
    rotateLeft(n);
  };

  /**
   * finds a node in the tree by a given value
   * @param {(string|number)} value
   * @returns {object} node
   */
  var find = function find(value) {
    var currentNode = rootNode;
    while (currentNode !== null) {
      if (value > currentNode.getValue()) {
        currentNode = currentNode.getRight();
      } else if (value < currentNode.getValue()) {
        currentNode = currentNode.getLeft();
      } else {
        return currentNode;
      }
    }
    return null;
  };

  /**
   * inserts a node by a given value into the tree
   * @param {(string|number)} value
   */
  var insert = function insert(value) {
    var balanceNode = function balanceNode(currentNode) {
      updateHeight(currentNode);
      var balance = calculateBalance(currentNode);
      if (balance > 1) {
        if (value < currentNode.getLeft().getValue()) {
          rotateRight(currentNode);
        } else {
          rotateLeftRight(currentNode);
        }
      } else if (balance < -1) {
        if (value > currentNode.getRight().getValue()) {
          rotateLeft(currentNode);
        } else {
          rotateRightLeft(currentNode);
        }
      }
    };

    var insertFn = function insertFn(currentNode) {
      if (currentNode === null) {
        rootNode = node(value);
        nodesCount += 1;
      } else if (value < currentNode.getValue()) {
        if (currentNode.getLeft() === null) {
          currentNode.setLeft(node(value, currentNode));
          nodesCount += 1;
        } else {
          insertFn(currentNode.getLeft());
        }
      } else if (value > currentNode.getValue()) {
        if (currentNode.getRight() === null) {
          currentNode.setRight(node(value, currentNode));
          nodesCount += 1;
        } else {
          insertFn(currentNode.getRight());
        }
      }
      balanceNode(currentNode);
    };

    insertFn(rootNode);
  };

  /**
   * removes a node by a given value from the tree
   * @param {(string|number)} value
   */
  var remove = function remove(value) {
    var balanceNode = function balanceNode(currentNode) {
      updateHeight(currentNode);
      var balance = calculateBalance(currentNode);
      if (balance > 1) {
        var left = currentNode.getLeft();
        if (left.getLeft() !== null) {
          rotateRight(currentNode);
        } else if (left.getRight() !== null) {
          rotateLeftRight(currentNode);
        }
      } else if (balance < -1) {
        var right = currentNode.getRight();
        if (right.getRight() !== null) {
          rotateLeft(currentNode);
        } else if (right.getLeft() !== null) {
          rotateRightLeft(currentNode);
        }
      }
    };

    var removeFn = function removeFn(val, currentNode) {
      if (currentNode !== null) {
        var left = currentNode.getLeft();
        var right = currentNode.getRight();
        if (val > currentNode.getValue()) {
          removeFn(val, right);
        } else if (val < currentNode.getValue()) {
          removeFn(val, left);
        } else {
          var parent = currentNode.getParent();
          if (right === null && left === null) {
            // remove a node with no children
            if (parent === null) {
              rootNode = null;
            } else if (currentNode.getValue() >= parent.getValue()) {
              parent.setRight(null);
            } else {
              parent.setLeft(null);
            }
            nodesCount -= 1;
          } else if (right === null) {
            // remove a node with a left child
            if (parent === null) {
              rootNode = left;
            } else if (currentNode.getValue() > parent.getValue()) {
              parent.setRight(left);
            } else {
              parent.setLeft(left);
            }
            left.setParent(parent);
            nodesCount -= 1;
          } else if (left === null) {
            // remove a node with a right child
            if (parent === null) {
              rootNode = right;
            } else if (currentNode.getValue() > parent.getValue()) {
              parent.setRight(right);
            } else {
              parent.setLeft(right);
            }
            right.setParent(parent);
            nodesCount -= 1;
          } else {
            // remove a node with two children
            var minRight = min(right);
            currentNode.setValue(minRight.getValue());
            removeFn(minRight.getValue(), minRight);
          }
        }
        balanceNode(currentNode);
      }
    };

    removeFn(value, rootNode);
  };

  /**
   * traverse the avl tree in-order (left-node-right)
   * @param {function} cb - called with each node value
   */
  var traverseInOrder = function traverseInOrder(cb) {
    var traverseInOrderFn = function traverseInOrderFn(currentNode) {
      if (currentNode !== null) {
        traverseInOrderFn(currentNode.getLeft());
        cb(currentNode);
        traverseInOrderFn(currentNode.getRight());
      }
    };
    traverseInOrderFn(rootNode);
  };

  /**
   * traverse the avl tree pre-order (node-left-right)
   * @param {function} cb - called with each node value
   */
  var traversePreOrder = function traversePreOrder(cb) {
    var traversePreOrderFn = function traversePreOrderFn(currentNode) {
      if (currentNode !== null) {
        cb(currentNode);
        traversePreOrderFn(currentNode.getLeft());
        traversePreOrderFn(currentNode.getRight());
      }
    };
    traversePreOrderFn(rootNode);
  };

  /**
   * traverse the avl tree post-order (left-right-node)
   * @param {function} cb - called with each node value
   */
  var traversePostOrder = function traversePostOrder(cb) {
    var traversePostOrderFn = function traversePostOrderFn(currentNode) {
      if (currentNode !== null) {
        traversePostOrderFn(currentNode.getLeft());
        traversePostOrderFn(currentNode.getRight());
        cb(currentNode);
      }
    };
    traversePostOrderFn(rootNode);
  };

  /**
   * traverse the avl tree
   * @param {function} cb - called with each node value
   * @param {string} type - 'inOrder' | 'preOrder' | 'postOrder'
   */
  var traverse = function traverse(cb, type) {
    switch (type) {
      case 'inOrder':
        traverseInOrder(cb);
        break;
      case 'preOrder':
        traversePreOrder(cb);
        break;
      case 'postOrder':
        traversePostOrder(cb);
        break;
      default:
        traverseInOrder(cb);
    }
  };

  /**
   * clears the tree
   */
  var clear = function clear() {
    rootNode = null;
    nodesCount = 0;
  };

  // avl tree api
  return {
    node: node,
    root: root,
    count: count,
    clear: clear,
    max: max,
    min: min,
    find: find,
    insert: insert,
    remove: remove,
    traverseInOrder: traverseInOrder,
    traversePreOrder: traversePreOrder,
    traversePostOrder: traversePostOrder,
    traverse: traverse
  };
};

exports.default = avlTree;