'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
/**
 * datastructures-js/binary-search-tree
 * @copyright 2018 Eyas Ranjous <eyas.ranjous@gmail.com>
 * @license MIT
 */

/**
 * binary tree node
 * @function
 */
var node = function node(v, p, l, r) {
  var value = v;
  var parent = p || null;
  var left = l || null;
  var right = r || null;

  /**
   * @param {object} value
   */
  var setValue = function setValue(val) {
    value = val;
  };

  /**
   * @returns {string|number}
   */
  var getValue = function getValue() {
    return value;
  };

  /**
   * @param {object} pr
   */
  var setParent = function setParent(pr) {
    parent = pr;
  };

  /**
   * @returns {object} node
   */
  var getParent = function getParent() {
    return parent;
  };

  /**
   * @param {BinaryNode} node
   */
  var setLeft = function setLeft(lf) {
    left = lf;
  };

  /**
   * @returns {object} node
   */
  var getLeft = function getLeft() {
    return left;
  };

  /**
   * @param {object} rg
   */
  var setRight = function setRight(rg) {
    right = rg;
  };

  /**
   * @returns {BinaryNode}
   */
  var getRight = function getRight() {
    return right;
  };

  // binary tree node api
  return {
    setValue: setValue,
    getValue: getValue,
    setParent: setParent,
    getParent: getParent,
    setRight: setRight,
    getRight: getRight,
    setLeft: setLeft,
    getLeft: getLeft
  };
};

/**
 * binary search tree
 * @function
 */
var binarySearchTree = function binarySearchTree() {
  var rootNode = null;
  var nodesCount = 0;

  /**
   * @returns {object} node
   */
  var root = function root() {
    return rootNode;
  };

  /**
   * @returns {number}
   */
  var count = function count() {
    return nodesCount;
  };

  /**
   * gets max value node in the tree
   * @returns {object} node
   */
  var max = function max(startingNode) {
    var currentNode = startingNode || rootNode;
    while (currentNode !== null && currentNode.getRight() !== null) {
      currentNode = currentNode.getRight();
    }
    return currentNode;
  };

  /**
   * gets min value node in the tree
   * @returns {object} node
   */
  var min = function min(startingNode) {
    var currentNode = startingNode || rootNode;
    while (currentNode !== null && currentNode.getLeft() !== null) {
      currentNode = currentNode.getLeft();
    }
    return currentNode;
  };

  /**
   * finds a node in the tree by a given value
   * @param {(string|number)} value
   * @returns {object} node
   */
  var find = function find(value) {
    var currentNode = rootNode;
    while (currentNode !== null) {
      if (value > currentNode.getValue()) {
        currentNode = currentNode.getRight();
      } else if (value < currentNode.getValue()) {
        currentNode = currentNode.getLeft();
      } else {
        return currentNode;
      }
    }
    return null;
  };

  /**
   * inserts a node by a given value into the tree
   * @param {(string|number)} value
   */
  var insert = function insert(value) {
    var insertFn = function insertFn(currentNode) {
      if (currentNode === null) {
        rootNode = node(value);
        nodesCount += 1;
      } else if (value < currentNode.getValue()) {
        if (currentNode.getLeft() === null) {
          currentNode.setLeft(node(value, currentNode));
          nodesCount += 1;
        } else {
          insertFn(currentNode.getLeft());
        }
      } else if (value > currentNode.getValue()) {
        if (currentNode.getRight() === null) {
          currentNode.setRight(node(value, currentNode));
          nodesCount += 1;
        } else {
          insertFn(currentNode.getRight());
        }
      }
    };
    insertFn(rootNode);
  };

  /**
   * removes a node by a given value from the tree
   * @param {(string|number)} value
   */
  var remove = function remove(value) {
    var removeFn = function removeFn(val, currentNode) {
      if (currentNode !== null) {
        var left = currentNode.getLeft();
        var right = currentNode.getRight();
        if (val > currentNode.getValue()) {
          removeFn(val, right);
        } else if (val < currentNode.getValue()) {
          removeFn(val, left);
        } else {
          var parent = currentNode.getParent();
          if (right === null && left === null) {
            // remove a node with no children
            if (parent === null) {
              rootNode = null;
            } else if (currentNode.getValue() >= parent.getValue()) {
              parent.setRight(null);
            } else {
              parent.setLeft(null);
            }
            nodesCount -= 1;
          } else if (right === null) {
            // remove a node with a left child
            if (parent === null) {
              rootNode = left;
            } else if (currentNode.getValue() > parent.getValue()) {
              parent.setRight(left);
            } else {
              parent.setLeft(left);
            }
            left.setParent(parent);
            nodesCount -= 1;
          } else if (left === null) {
            // remove a node with a right child
            if (parent === null) {
              rootNode = right;
            } else if (currentNode.getValue() > parent.getValue()) {
              parent.setRight(right);
            } else {
              parent.setLeft(right);
            }
            right.setParent(parent);
            nodesCount -= 1;
          } else {
            // remove a node with two children
            var minRight = min(right);
            currentNode.setValue(minRight.getValue());
            removeFn(minRight.getValue(), minRight);
          }
        }
      }
    };
    removeFn(value, rootNode);
  };

  /**
   * traverse the binary tree in-order (left-parent-right)
   * @param {function} cb - called with each node value
   */
  var traverseInOrder = function traverseInOrder(cb) {
    var traverseInOrderFn = function traverseInOrderFn(currentNode) {
      if (currentNode !== null) {
        traverseInOrderFn(currentNode.getLeft());
        cb(currentNode);
        traverseInOrderFn(currentNode.getRight());
      }
    };
    traverseInOrderFn(rootNode);
  };

  /**
   * traverse the binary tree pre-order (parent-left-right)
   * @param {function} cb - called with each node value
   */
  var traversePreOrder = function traversePreOrder(cb) {
    var traversePreOrderFn = function traversePreOrderFn(currentNode) {
      if (currentNode !== null) {
        cb(currentNode);
        traversePreOrderFn(currentNode.getLeft());
        traversePreOrderFn(currentNode.getRight());
      }
    };
    traversePreOrderFn(rootNode);
  };

  /**
   * traverse the binary tree post-order (left-right-parent)
   * @param {function} cb - called with each node value
   */
  var traversePostOrder = function traversePostOrder(cb) {
    var traversePostOrderFn = function traversePostOrderFn(currentNode) {
      if (currentNode !== null) {
        traversePostOrderFn(currentNode.getLeft());
        traversePostOrderFn(currentNode.getRight());
        cb(currentNode);
      }
    };
    traversePostOrderFn(rootNode);
  };

  /**
   * traverse the binary tree
   * @param {function} cb - called with each node value
   * @param {string} type - 'inOrder' | 'preOrder' | 'postOrder'
   */
  var traverse = function traverse(cb, type) {
    switch (type) {
      case 'inOrder':
        traverseInOrder(cb);
        break;
      case 'preOrder':
        traversePreOrder(cb);
        break;
      case 'postOrder':
        traversePostOrder(cb);
        break;
      default:
        traverseInOrder(cb);
    }
  };

  /**
   * clears the tree
   */
  var clear = function clear() {
    rootNode = null;
    nodesCount = 0;
  };

  // binary tree api
  return {
    node: node,
    root: root,
    count: count,
    clear: clear,
    max: max,
    min: min,
    find: find,
    insert: insert,
    remove: remove,
    traverseInOrder: traverseInOrder,
    traversePreOrder: traversePreOrder,
    traversePostOrder: traversePostOrder,
    traverse: traverse
  };
};

exports.default = binarySearchTree;