'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _queue = require('./queue');

var _queue2 = _interopRequireDefault(_queue);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * graph vertex
 * @function
 */
var vertex = function vertex(k, v) {
  var key = k;
  var value = v;

  /**
   * @returns {(string|number)}
   */
  var getKey = function getKey() {
    return key;
  };

  /**
   * @param {object} val
   */
  var setValue = function setValue(val) {
    value = val;
  };

  /**
   * @returns {object}
   */
  var getValue = function getValue() {
    return value;
  };

  // vertex api
  return {
    getKey: getKey,
    setValue: setValue,
    getValue: getValue
  };
};

/**
 * graph & directed graph
 * @function
 */
/**
 * datastructures-js/graph
 * @copyright 2018 Eyas Ranjous <eyas.ranjous@gmail.com>
 * @license MIT
 */
var graph = function graph(options) {
  var vertices = {};
  var edges = {};
  var verticesCount = 0;

  var _ref = options && options || false,
      directed = _ref.directed;

  var self = {};

  /**
   * add a vertex to the graph
   * @param {(string|number)} key
   * @param {object} value
   */
  self.addVertex = function (key, value) {
    if (!vertices[key]) {
      vertices[key] = vertex(key, value);
      edges[key] = {};
      verticesCount += 1;
    }
  };

  /**
   * removes a vertex from the graph
   * @param {(string|number} key
   */
  self.removeVertex = function (key) {
    if (vertices[key]) {
      vertices[key] = undefined;
      verticesCount -= 1;
    }
  };

  /**
   * checks if the graph has a vertex
   * @param {(string|number)} key
   * @returns {boolean}
   */
  self.hasVertex = function (key) {
    return !!vertices[key];
  };

  /**
   * gets the count of the vertices in the graph
   * @returns {number}
   */
  self.countVertices = function () {
    return verticesCount;
  };

  /**
   * adds an edge between two existing vertices
   * @param {(string|number)} key1 - first vertext key
   * @param {(string|number)} key2 - second vertex key
   * @param {number} the numeric weight of the edge
   */
  self.addEdge = function (key1, key2, weight) {
    if (self.hasVertex(key1) && self.hasVertex(key2)) {
      var w = +weight || 0;
      edges[key1][key2] = w;
      if (!directed) {
        edges[key2][key1] = w;
      }
    }
  };

  /**
   * checks if the graph has an edge between two existing vertices
   * @param {(string|number)} key1
   * @param {(string|number)} key2
   * @returns {boolean}
   */
  self.hasEdge = function (key1, key2) {
    if (self.hasVertex(key1) && self.hasVertex(key2)) {
      if (directed && edges[key1][key2] >= 0) {
        return true;
      } else if (edges[key1][key2] >= 0 && edges[key2][key1] >= 0) {
        return true;
      }
    }
    return false;
  };

  /**
   * gets the weight of the edge between two vertices
   * @returns {number|null}
   */
  self.getWeight = function (key1, key2) {
    if (self.hasVertex(key1) && key1 === key2) {
      return 0;
    } else if (self.hasEdge(key1, key2)) {
      return edges[key1][key2];
    }
    return null;
  };

  /**
   * gets the connected vertices to a vertex
   * @returns {number|null}
   */
  self.getEdges = function (key) {
    return edges[key] || {};
  };

  /**
   * removes an existing edge between two vertices
   * @param {(string|number)} key1
   * @param {(string|number)} key2
   */
  self.removeEdge = function (key1, key2) {
    if (self.hasEdge(key1, key2)) {
      edges[key1][key2] = -1;
      edges[key2][key1] = -1;
    }
  };

  /**
   * traverse the graph using depth-first approach from a starting vertex
   * @param {(string|number)} key - starting vertex key
   * @param {function} cb - called with each vertext object
   */
  var dfsTraverse = function dfsTraverse(key, cb) {
    var visited = {};
    var traverse = function traverse(currentKey) {
      if (self.hasVertex(currentKey) && !visited[currentKey]) {
        cb(vertices[currentKey]);
        visited[currentKey] = true;
        Object.keys(edges[currentKey]).forEach(function (k) {
          if (self.hasEdge(currentKey, k) && !visited[k]) {
            traverse(k);
          }
        });
      }
    };
    traverse(key);
  };

  /**
   * traverse the graph using breadth-first approach from a starting vertex
   * @param {(string|number)} key - starting vertex key
   * @param {function} cb - called with each vertext object
   */
  var bfsTraverse = function bfsTraverse(key, cb) {
    if (self.hasVertex(key)) {
      (function () {
        var queue = (0, _queue2.default)();
        var visited = {};
        queue.enqueue(key);
        visited[key] = true;

        var _loop = function _loop() {
          var k1 = queue.dequeue();
          cb(vertices[k1]);
          Object.keys(edges[k1]).forEach(function (k2) {
            if (self.hasEdge(k1, k2) && !visited[k2]) {
              queue.enqueue(k2);
              visited[k2] = true;
            }
          });
        };

        while (!queue.isEmpty()) {
          _loop();
        }
      })();
    }
  };

  /**
   * traverse the graph using using a traversal algorithm
   * @param {(string|number)} key - starting vertex key
   * @param {function} cb - called with each vertex value
   * @param {string} - traversal type, default is bfs
   */
  self.traverse = function (key, cb, type) {
    switch (type) {
      case 'dfs':
        dfsTraverse(key, cb);
        break;
      case 'bfs':
        bfsTraverse(key, cb);
        break;
      default:
        bfsTraverse(key, cb);
        break;
    }
  };

  /**
   * clears the graph
   */
  self.clear = function () {
    vertices = {};
    edges = {};
    verticesCount = 0;
  };

  return self;
};

exports.default = graph;