"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
/**
 * datastructures-js/linked-list
 * @copyright 2018 Eyas Ranjous <eyas.ranjous@gmail.com>
 * @license MIT
 */

/**
 * Linked List Node
 * @function
 */
var node = function node(val, nex) {
  var value = val;
  var next = nex;

  /**
   * @param {string|number} v
   */
  var setValue = function setValue(v) {
    value = v;
  };

  /**
   * @return {string|number}
   */
  var getValue = function getValue() {
    return value;
  };

  /**
   * @param {object} - node
   */
  var setNext = function setNext(n) {
    next = n;
  };

  /**
   * @return {object} - node
   */
  var getNext = function getNext() {
    return next || null;
  };

  // linkedList node API
  return {
    setValue: setValue,
    getValue: getValue,
    setNext: setNext,
    getNext: getNext
  };
};

/**
 * Linked List
 * @function
 */
var linkedList = function linkedList() {
  var headNode = null;
  var nodesCount = 0;

  /**
   * @returns {object} - node
   */
  var head = function head() {
    return headNode;
  };

  /**
   * @returns {number}
   */
  var count = function count() {
    return nodesCount;
  };

  /**
   * @param {string|number} value
   * @returns {object} - node
   */
  var find = function find(value) {
    var current = headNode;
    while (current !== null) {
      if (current.getValue() === value) {
        return current;
      }
      current = current.getNext();
    }
    return current;
  };

  /**
   * @param {string|number} value
   */
  var addFirst = function addFirst(value) {
    if (headNode === null) {
      headNode = node(value);
    } else {
      headNode = node(value, headNode);
    }
    nodesCount += 1;
  };

  /**
   * @param {string|number} value
   */
  var addLast = function addLast(value) {
    if (headNode === null) {
      headNode = node(value);
    } else {
      var current = headNode;
      while (current.getNext() !== null) {
        current = current.getNext();
      }
      current.setNext(node(value));
    }
    nodesCount += 1;
  };

  /**
   * adds a new node after an existing node
   * @param {string|number} value
   * @param {string|number} newValue
   * @throws {Error}
   */
  var addAfter = function addAfter(value, newValue) {
    var current = headNode;
    while (current !== null) {
      if (current.getValue() === value) {
        current.setNext(node(newValue, current.getNext()));
        nodesCount += 1;
        break;
      } else {
        current = current.getNext();
      }
    }
    if (current === null) {
      throw new Error("node " + value + " not found");
    }
  };

  /**
   * @public
   * adds a new node before an existing node
   * @param {(string|number)} value
   * @param {(string|number)} newValue
   * @throws {Error}
   */
  var addBefore = function addBefore(value, newValue) {
    var prev = null;
    var current = headNode;
    while (current !== null) {
      if (current.getValue() === value) {
        if (prev === null) {
          addFirst(newValue);
        } else {
          prev.setNext(node(newValue, current));
          nodesCount += 1;
        }
        break;
      } else {
        prev = current;
        current = current.getNext();
      }
    }
    if (current === null) {
      throw new Error("node " + value + " not found");
    }
  };

  /**
   * removes the first node
   */
  var removeFirst = function removeFirst() {
    if (headNode !== null) {
      if (headNode.getNext() === null) {
        headNode = null;
      } else {
        headNode = headNode.getNext();
      }
      nodesCount -= 1;
    }
  };

  /**
   * removes the last node
   */
  var removeLast = function removeLast() {
    var prev = null;
    var current = headNode;
    while (current.getNext() !== null) {
      prev = current;
      current = current.getNext();
    }
    if (prev === null) {
      removeFirst();
    } else {
      prev.setNext(null);
      nodesCount -= 1;
    }
  };

  /**
   * removes a node by its value
   * @param {(string|number)} value
   */
  var remove = function remove(value) {
    var prev = null;
    var current = headNode;
    while (current !== null) {
      if (current.getValue() === value) {
        if (prev === null) {
          removeFirst();
        } else {
          prev.setNext(current.getNext());
          nodesCount -= 1;
        }
        break;
      } else {
        prev = current;
        current = current.getNext();
      }
    }
  };

  /**
   * traverse the linkedlist from a beginning to end
   * @param {function} cb - called with node in the linked list
   */
  var traverse = function traverse(cb) {
    var current = headNode;
    while (current !== null) {
      cb(current);
      current = current.getNext();
    }
  };

  /**
   * @returns {array}
   */
  var toArray = function toArray() {
    var arr = [];
    traverse(function (n) {
      return arr.push(n.getValue());
    });
    return arr;
  };

  /**
   * clears the linked list
   */
  var clear = function clear() {
    headNode = null;
    nodesCount = 0;
  };

  // linkedList API
  return {
    node: node,
    head: head,
    count: count,
    find: find,
    addFirst: addFirst,
    addLast: addLast,
    addBefore: addBefore,
    addAfter: addAfter,
    remove: remove,
    removeFirst: removeFirst,
    removeLast: removeLast,
    traverse: traverse,
    toArray: toArray,
    clear: clear
  };
};

exports.default = linkedList;