'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
/**
 * datastructures-js/trie
 * @copyright 2018 Eyas Ranjous <eyas.ranjous@gmail.com>
 * @license MIT
 */

/**
 * @throws {Error}
 */
var validate = function validate(str) {
  if (typeof str !== 'string') {
    throw new Error(str + ' is not a string');
  }
};

/**
 * Trie Node
 * @function
 */
var node = function node(ch) {
  var char = ch;
  var endOfWord = false;
  var parent = null;
  var children = {};
  var childrenCount = 0;

  validate(ch);

  /**
   * @return {string}
   */
  var getChar = function getChar() {
    return char;
  };

  /**
   * @param {object} p
   */
  var setParent = function setParent(p) {
    parent = p;
  };

  /**
   * @return {object} - node
   */
  var getParent = function getParent() {
    return parent;
  };

  /**
   * @param {boolean}
   */
  var setEndOfWord = function setEndOfWord(eow) {
    endOfWord = eow;
  };

  /**
   * @returns {boolean}
   */
  var isEndOfWord = function isEndOfWord() {
    return endOfWord;
  };

  /**
   * @param {node} child
   */
  var addChild = function addChild(child) {
    children[child.getChar()] = child;
    childrenCount += 1;
  };

  /**
   * @param {node} child
   */
  var removeChild = function removeChild(child) {
    children[child.getChar()] = null;
    childrenCount -= 1;
  };

  /**
   * @returns {node} - child
   */
  var getChild = function getChild(c) {
    return children[c] || null;
  };

  /**
   * @returns {object}
   */
  var getChildren = function getChildren() {
    return children;
  };

  /**
   * @returns {number}
   */
  var countChildren = function countChildren() {
    return childrenCount;
  };

  // trie node api
  return {
    getChar: getChar,
    setParent: setParent,
    getParent: getParent,
    setEndOfWord: setEndOfWord,
    isEndOfWord: isEndOfWord,
    addChild: addChild,
    removeChild: removeChild,
    getChild: getChild,
    getChildren: getChildren,
    countChildren: countChildren
  };
};

/**
 * Trie
 * @function
 */
var trie = function trie() {
  var nodesCount = 1;
  var wordsCount = 0;
  var rootNode = node('');

  /**
   * @returns {number}
   */
  var countNodes = function countNodes() {
    return nodesCount;
  };

  /**
   * @returns {number}
   */
  var countWords = function countWords() {
    return wordsCount;
  };

  /**
   * finds a word in the trie
   * @param {string} word
   * @returns {object} - node
   */
  var search = function search(word) {
    validate(word);
    var currentNode = rootNode;
    for (var i = 0; i < word.length; i += 1) {
      var child = currentNode.getChild(word[i]);
      if (child === null) {
        return null;
      }
      currentNode = child;
    }
    if (currentNode.isEndOfWord()) {
      return currentNode;
    }
    return null;
  };

  /**
   * inserts a word into the trie
   * @param {string} word
   */
  var insert = function insert(word) {
    validate(word);
    var currentNode = rootNode;
    for (var i = 0; i < word.length; i += 1) {
      if (currentNode.getChild(word[i]) === null) {
        var child = node(word[i]);
        child.setParent(currentNode);
        currentNode.addChild(child);
        nodesCount += 1;
      }
      currentNode = currentNode.getChild(word[i]);
    }
    if (currentNode.getChar() !== '') {
      currentNode.setEndOfWord(true);
      wordsCount += 1;
    }
  };

  /**
   * removes a word from the trie
   * @param {string} word
   */
  var remove = function remove(word) {
    var currentNode = search(word);
    if (currentNode !== null && currentNode.getChar() !== '') {
      if (currentNode.countChildren() > 0) {
        currentNode.setEndOfWord(false);
      } else {
        while (currentNode.getParent() !== null) {
          if (currentNode.countChildren() === 0) {
            currentNode.getParent().removeChild(currentNode);
            nodesCount -= 1;
          }
          currentNode = currentNode.getParent();
        }
      }
      wordsCount -= 1;
    }
  };

  /**
   * depth-first traverse the trie from root node
   * @param {function} cb
   */
  var traverse = function traverse(cb) {
    var word = '';
    var traverseFn = function traverseFn(currentNode) {
      if (currentNode.isEndOfWord()) {
        cb(word);
      }
      var children = currentNode.getChildren();
      Object.keys(children).forEach(function (char) {
        var child = children[char];
        word += child.getChar();
        traverseFn(child); // depth-first traverse
        word = word.substr(0, word.length - 1); // word backward tracking
      });
    };
    return traverseFn(rootNode);
  };

  /**
   * clears the trie
   */
  var clear = function clear() {
    nodesCount = 1;
    wordsCount = 0;
    rootNode = node('');
  };

  // trie api
  return {
    node: node,
    countNodes: countNodes,
    countWords: countWords,
    search: search,
    insert: insert,
    remove: remove,
    traverse: traverse,
    clear: clear
  };
};

exports.default = trie;